"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const tsutils_1 = require("tsutils");
const total_functions_1 = require("total-functions");
/**
 * An ESLint rule to ban unsafe assignment and declarations.
 */
const noUnsafeAssignment = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Bans unsafe type assertions.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringCallExpressionReadonlyToMutable: "Passing a readonly type to a function that expects a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringAssignmentExpressionReadonlyToMutable: "Assigning a readonly type to a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringVariableDeclarationReadonlyToMutable: "Using a readonly type to initialize a mutable type can lead to unexpected mutation in the readonly value.",
        },
        schema: [],
    },
    create: (context) => {
        // eslint-disable-next-line total-functions/no-unsafe-assignment
        const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const isUnsafeAssignment = (destinationType, sourceType, seenTypes = []
        // eslint-disable-next-line sonarjs/cognitive-complexity
        ) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (
            // TODO what if the source type is a union that contains a mix of mutable and readonly types?
            !sourceType.isUnion() &&
                // TODO what if the destination type is a union that contains a mix of mutable and readonly types?
                // If the source type is mutable and corresponds to the (safe) mutable destination type, this will still be flagged
                // as unsafe (due to the readonly portion of the destination type union).
                !destinationType.isUnion() &&
                tsutils_1.isObjectType(sourceType) &&
                tsutils_1.isObjectType(destinationType)) {
                // TODO this needs to check arrays too (the array itself being readonly AND its generic type being readonly)
                // TODO how to handle tuples?
                return destinationType.getProperties().some((destinationProperty) => {
                    const destinationPropIsReadonly = tsutils_1.isPropertyReadonlyInType(destinationType, destinationProperty.getEscapedName(), checker);
                    const sourceProperty = sourceType.getProperty(destinationProperty.name);
                    const sourcePropIsReadonly = sourceProperty !== undefined &&
                        tsutils_1.isPropertyReadonlyInType(sourceType, sourceProperty.getEscapedName(), checker);
                    const destinationPropertyDeclarations = destinationProperty.getDeclarations() || [];
                    // TODO: How to choose declaration when there are multiple? `checker.getResolvedSignature()`?
                    const destinationPropertyDeclaration = destinationPropertyDeclarations.length > 1
                        ? undefined
                        : total_functions_1.get(destinationPropertyDeclarations, 0);
                    const destinationPropertyType = destinationPropertyDeclaration !== undefined
                        ? checker.getTypeAtLocation(destinationPropertyDeclaration)
                        : undefined;
                    const sourcePropertyDeclarations = sourceProperty === undefined
                        ? []
                        : sourceProperty.getDeclarations() || [];
                    // TODO: How to choose declaration when there are multiple? `checker.getResolvedSignature()`?
                    const sourcePropertyDeclaration = sourcePropertyDeclarations.length > 1
                        ? undefined
                        : total_functions_1.get(sourcePropertyDeclarations, 0);
                    const sourcePropertyType = sourcePropertyDeclaration !== undefined
                        ? checker.getTypeAtLocation(sourcePropertyDeclaration)
                        : undefined;
                    const isUnsafeAssignmentRecursively = destinationPropertyType !== undefined &&
                        sourcePropertyType !== undefined &&
                        // Try to avoid infinite recursion...
                        seenTypes.every((t) => t.destinationType !== destinationPropertyType &&
                            t.sourceType !== sourcePropertyType) &&
                        isUnsafeAssignment(destinationPropertyType, sourcePropertyType, seenTypes.concat([{ destinationType, sourceType }]));
                    const assigningReadonlyToMutable = sourcePropIsReadonly && !destinationPropIsReadonly;
                    return assigningReadonlyToMutable || isUnsafeAssignmentRecursively;
                });
            }
            else {
                return false;
            }
        };
        return {
            // eslint-disable-next-line functional/no-return-void
            VariableDeclaration: (node) => {
                // eslint-disable-next-line functional/no-expression-statement
                node.declarations.forEach((declaration) => {
                    const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(declaration.id);
                    const rightTsNode = declaration.init !== null
                        ? parserServices.esTreeNodeToTSNodeMap.get(declaration.init)
                        : undefined;
                    const leftType = checker.getTypeAtLocation(leftTsNode);
                    const rightType = rightTsNode !== undefined
                        ? checker.getTypeAtLocation(rightTsNode)
                        : undefined;
                    // eslint-disable-next-line functional/no-conditional-statement
                    if (declaration.init !== null &&
                        rightType !== undefined &&
                        // object expressions are allowed because we won't retain a reference to the object to get out of sync.
                        // TODO but what about properties in the object literal that are references to values we _do_ retain a reference to?
                        declaration.init.type !== experimental_utils_1.AST_NODE_TYPES.ObjectExpression &&
                        isUnsafeAssignment(leftType, rightType)) {
                        // eslint-disable-next-line functional/no-expression-statement
                        context.report({
                            node: node,
                            messageId: "errorStringVariableDeclarationReadonlyToMutable",
                        });
                    }
                });
            },
            // eslint-disable-next-line functional/no-return-void
            AssignmentExpression: (node) => {
                const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.left);
                const rightTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.right);
                const leftType = checker.getTypeAtLocation(leftTsNode);
                const rightType = checker.getTypeAtLocation(rightTsNode);
                // eslint-disable-next-line functional/no-conditional-statement
                if (
                // object expressions are allowed because we won't retain a reference to the object to get out of sync.
                // TODO but what about properties in the object literal that are references to values we _do_ retain a reference to?
                node.right.type !== experimental_utils_1.AST_NODE_TYPES.ObjectExpression &&
                    isUnsafeAssignment(leftType, rightType)) {
                    // eslint-disable-next-line functional/no-expression-statement
                    context.report({
                        node: node,
                        messageId: "errorStringAssignmentExpressionReadonlyToMutable",
                    });
                }
            },
            // TODO
            // eslint-disable-next-line functional/no-return-void
            // ReturnStatement: (node): void => {
            // },
            // // eslint-disable-next-line functional/no-return-void
            // ArrowFunctionExpression: (node): void => {
            // },
            // eslint-disable-next-line functional/no-return-void, sonarjs/cognitive-complexity
            CallExpression: (node) => {
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node.callee);
                const tsType = checker.getTypeAtLocation(tsNode);
                const signatures = tsType.getCallSignatures();
                // TODO: if a function has more than one signature, how do we know which
                // one applies for this call? `checker.getResolvedSignature()`?
                const signature = total_functions_1.get(signatures, 0);
                // eslint-disable-next-line functional/no-conditional-statement
                if (signatures.length === 1 && signature !== undefined) {
                    // eslint-disable-next-line functional/no-expression-statement
                    signature.parameters.forEach((parameter, i) => {
                        const paramType = checker.getTypeOfSymbolAtLocation(parameter, tsNode);
                        // This is the argument that corresponds to the current parameter.
                        const argument = total_functions_1.get(node.arguments, i);
                        const argumentTsNode = argument !== undefined
                            ? parserServices.esTreeNodeToTSNodeMap.get(argument)
                            : undefined;
                        const argumentType = argumentTsNode !== undefined
                            ? checker.getTypeAtLocation(argumentTsNode)
                            : undefined;
                        // eslint-disable-next-line functional/no-conditional-statement
                        if (argument !== undefined &&
                            argumentType !== undefined &&
                            // object expressions are allowed because we won't retain a reference to the object to get out of sync.
                            // TODO but what about properties in the object literal that are references to values we _do_ retain a reference to?
                            argument.type !== experimental_utils_1.AST_NODE_TYPES.ObjectExpression &&
                            isUnsafeAssignment(paramType, argumentType)) {
                            // eslint-disable-next-line functional/no-expression-statement
                            context.report({
                                node: argument,
                                messageId: "errorStringCallExpressionReadonlyToMutable",
                            });
                        }
                    });
                }
            },
        };
    },
};
exports.default = noUnsafeAssignment;
